import os
import sys
import time
import shutil
import tempfile
import functools
import subprocess

import zope.interface
from twisted.python import usage, log
from twisted.protocols.basic import LineReceiver
from twisted.internet import defer, reactor, stdio

import txtorcon
from carml import util

_log = functools.partial(log.msg, system='carml')


def dump_xplanet_files(cfg, all, arc_file, file, follow, state):
    print("dumping", arc_file, file)
    unique_routers = set()
    routers_in_streams = set()
    if all:
        map(unique_routers.add, state.routers.values())

    else:
        # this method does only active routers (i.e. in at least
        # one of your streams) and colours guards as green
        for circ in state.circuits.values():
            map(unique_routers.add, circ.path)
        map(unique_routers.add, state.entry_guards.values())

        # cache any router that's used by any current circuit right now
        for circ in state.circuits.values():
            map(routers_in_streams.add, circ.path)

        if False:
            map(unique_routers.add, state.guards.values())

    header = '## Auto-generated "{}" by carml'.format(time.asctime())
    if arc_file is not None:
        arc_file.write(header + '\n')
        arc_file.write('## format: lat0 lng0 lat1 lng1\n\n')

        def gen_colors(r, g, b):
            while True:
                yield '0x%02x%02x%02x' % (r, g, b)
                r = int(r * 0.65)
                g = int(g * 0.65)
                b = int(b * 0.65)

        def gen_start_colors():
            while True:
                yield 0xb5, 0x89, 0x00
                yield 0xcb, 0x4b, 0x16
                yield 0xdc, 0x32, 0x2f
                yield 0xd3, 0x36, 0x82
                yield 0x6c, 0x71, 0xc4
                yield 0x26, 0x8b, 0xd2
                yield 0x2a, 0xa1, 0x98
                yield 0x85, 0x99, 0x00

        start_color = gen_start_colors()
        for circ in state.circuits.values():
            arc_file.write('## circuit {}\n'.format(circ.id))
            arc_colors = gen_colors(*start_color.next())
            for (i, link) in enumerate(circ.path[:-1]):
                nxt = circ.path[i + 1]
                if link.location.latlng[0] and nxt.location.latlng[0]:
                    arc_file.write('{} {} '.format(*link.location.latlng))
                    arc_file.write('{} {} '.format(*nxt.location.latlng))
                    arc_file.write('color={} thickness=2 # {}->{}\n'.format(arc_colors.next(), link.id_hex, nxt.id_hex))

    markerfile = file
    markerfile.write(header + '\n')
    markerfile.write('## {} unique routers\n'.format(len(unique_routers)))
    markerfile.write('## format: lat lng "name-or-hash" # hex-id\n\n')

    misses = 0
    for router in unique_routers:
        lat, lng = router.location.latlng
        if lat is not None and lng is not None:
            color = 'purple'
            if router.id_hex in state.entry_guards:
                color = 'red'
            if router in routers_in_streams:
                color = 'green'
            if color != 'green':
                markerfile.write('%02.5f %02.5f color=%s # %s\n' % (lat, lng, color, router.id_hex))

        else:
            markerfile.write('# unknown location: %s (%s)\n' % (router.unique_name, router.id_hex))
            misses += 1

    for router in routers_in_streams:
        lat, lng = router.location.latlng
        if lat and lng:
            markerfile.write('%02.5f %02.5f color=green # %s %s\n' % (lat, lng, router.unique_name, router.id_hex))

    if False:
        for stream in state.streams.values():
            for (idx, router) in enumerate(stream.circuit.path):
                lat, lng = router.location.latlng
                if lat and lng:
                    markerfile.write('%02.5f %02.5f "%d:%s" color=green # %s\n' % (lat, lng, idx, router.unique_name, router.id_hex))

    if not cfg.quiet:
        if misses == len(unique_routers):
            print('NOTE: it seems NO routers had location information.')
            print("Things to try:")
            print(" * install MaxMind IP geolocation database.")
            print(" * pip install geoip")
        else:
            if misses == 0:
                print('All routers had location information.')
            else:
                print('%d (%2.0f%%) routers with no geoip information.' % (misses, (float(misses) / len(unique_routers)) * 100.0))


class CircuitListener(txtorcon.CircuitListenerMixin):
    next_defer = None

    def _trigger_event(self, circuit, **kw):
        print(circuit)
        if self.next_defer:
            nd = self.next_defer
            self.next_defer = None
            nd.callback(circuit)

    circuit_failed = _trigger_event
    circuit_closed = _trigger_event
    circuit_built = _trigger_event


def generate_circuit_builds(listener):
    def generator():
        while True:
            listener.next_defer = defer.Deferred()
            yield listener.next_defer
    return generator()


@defer.inlineCallbacks
def continuously_update_xplanet(cfg, all, arc_file, file, follow, state):
    listener = CircuitListener()
    gen = generate_circuit_builds(listener)
    state.add_circuit_listener(listener)

    tmpdir = tempfile.mkdtemp()
    reactor.addSystemEventTrigger('before', 'shutdown',
                                  functools.partial(shutil.rmtree, tmpdir))

    marker_fname = os.path.join(tmpdir, 'xplanet_markers')
    arcs_fname = os.path.join(tmpdir, 'xplanet_arcs')
    cfg_fname = os.path.join(tmpdir, 'xplanet-config')

    with open(cfg_fname, 'w') as f:
        f.write('''[earth]\n"Earth"\nmarker_file=%s\narc_file=%s\n''' % (marker_fname, arcs_fname))

    cmd = ['xplanet',
           '-num_times', '1',
           '-projection', 'rectangular',
           '-config', cfg_fname,
           ]
    os.chdir(tmpdir)
    while True:
        file = open(marker_fname, 'w')
        arc_file = open(arcs_fname, 'w')
        dump_xplanet_files(cfg, all, arc_file, file, follow, state)
        if not cfg.quiet:
            print(' '.join(cmd), subprocess.check_output(cmd))

        if not follow:
            return

        circ = yield gen.next()
        print(circ)


@defer.inlineCallbacks
def run(reactor, cfg, tor, all, execute, follow, arc_file, file):

    state = yield tor.create_state()
    if follow or execute:
        yield continuously_update_xplanet(cfg, all, arc_file, file, follow, state)
    else:
        dump_xplanet_files(cfg, all, arc_file, file, follow, state)



'''
Utility functions.

Probably some of these could be replaced by a library or two,
especially the progress-bar hackery. The colour stuff is also a little
wonky.
'''

from __future__ import print_function

import datetime
import functools
from six import unichr

import colors


class NoColor(object):
    '''
    Replaces color methods with no-ops if we decide to turn off
    colored output.
    '''

    def __init__(self):
        self._overrides = [x for x in dir(colors) if not x.startswith('__')]

    def __getattr__(self, name):
        '''
        All the "things" in the colors library are methods; so, if we
        didn't ask for a magic-python method, we return the "noop"
        function which doesn't do anything to your string. The colors
        library would add ANSI commands around it.
        '''
        if name in self._overrides:
            return lambda x: x
        return self.__dict__[name]


def turn_off_color():
    "Here be the wonky bits..."
    global colors
    colors = NoColor()


def pretty_progress(percent, size=10, ascii=False):
    """
    Displays a unicode or ascii based progress bar of a certain
    length. Should we just depend on a library instead?
    """

    curr = int(percent / 100.0 * size)
    part = (percent / (100.0 / size)) - curr

    if ascii:
        part = int(part * 4)
        part = '.oO%'[part]
        block_chr = '#'

    else:
        block_chr = u'\u2588'
        # there are 8 unicode characters for vertical-bars/horiz-bars
        part = int(part * 8)

        # unicode 0x2581 -> 2589 are vertical bar chunks, like rainbarf uses
        # and following are narrow -> wider bars
        part = unichr(0x258f - part)  # for smooth bar
        # part = unichr(0x2581 + part) # for neater-looking thing

    # hack for 100+ full so we don't print extra really-narrow/high bar
    if percent >= 100.0:
        part = ''
    curr = int(curr)
    return '[%s%s%s]' % ((block_chr * curr), part, (' ' * (size - curr - 1)))


def wrap(text, width, prefix=''):
    """
    Simple word-wrapping thing. Might be worth considering a
    dependency on 'textwrap' if this gets longer than a few lines ;)
    """
    words = text.split()
    lines = []
    while len(words):
        line = prefix
        while len(words) and len(line) < width:
            word = words.pop(0)
            line += ' ' + word
        lines.append(line)
    return '\n'.join(lines)


def format_net_location(loc, verbose_asn=False):
    rtn = '(%s ' % loc.ip
    comma = False
    if loc.asn:
        if verbose_asn:
            rtn += loc.asn
        else:
            rtn += loc.asn.split()[0]
        comma = True
    if loc.countrycode:
        if comma:
            rtn += ', '
        rtn += loc.countrycode
        comma = True
    if loc.city and loc.city[0]:
        if comma:
            rtn += ', '
        rtn += ','.join([x.decode('utf8', 'replace') for x in loc.city])
        comma = True
    return rtn.strip() + ')'


def nice_router_name(router, color=True):
    """
    returns a router name with ~ at the front if it's not a named router
    """
    green = str
    italic = str
    if color:
        green = colors.green
        italic = colors.italic
    if router.name_is_unique:
        return green(router.name)
    return italic('~%s' % router.name)


def dump_circuits(state, verbose, show_countries=False):
    print('  %-4s | %-5s | %-42s | %-8s | %-12s' % ('ID', 'Age', 'Path (router names, ~ means no Named flag)', 'State', 'Purpose'))
    print(' ------+-------+' + ('-' * 44) + '+' + (10 * '-') + '+' + (12 * '-'))
    circuits = state.circuits.values()
    circuits.sort(lambda a, b: cmp(a.id, b.id))
    now = datetime.datetime.utcnow()
    for circ in circuits:
        path = '->'.join([nice_router_name(x) for x in circ.path])
        plain_router_name = functools.partial(nice_router_name, color=False)
        plain_path = '->'.join([plain_router_name(x) for x in circ.path])
        real_len = len(plain_path)
        if real_len > 42:
            # revert to uncoloured path since we don't know where ANSI controls are
            path = plain_path[:(42 - 13)] + '...' + plain_path[-10:]
        elif len(path) == 0:
            path = '(empty)'
        else:
            path = path + ((42 - real_len) * ' ')
        age = circ.age(now)
        if age > 300:
            age = '%2dmin' % (age / 60.0)
        else:
            age = '%ds' % age

        # path is already padded to 42 chars, as it contains ANSI controls
        print(colors.bold('  %4d | %5s | %s | %-8s | %-12s' % (circ.id, age, path, circ.state, circ.purpose)))
        # print str(circ.flags)
        if show_countries:
            print(' ' * 17, '->'.join(map(lambda x: x.location.countrycode, circ.path)))
        if verbose:
            padding = ' ' * 17
            print(' ' * 8, ', '.join([(str(k) + '=' + str(v)) for (k, v) in circ.flags.items()]))
            for router in circ.path:
                if router.ip != 'unknown':
                    print(padding, '%s=%s' % (router.name,
                                              format_net_location(router.location)))
    print(' ------+-------+' + ('-' * 44) + '+' + (10 * '-') + '+' + (12 * '-'))



from __future__ import print_function
import sys
import time
import functools

import zope.interface
from twisted.python import usage, log
from twisted.plugin import IPlugin
from twisted.internet import reactor
from twisted.internet import defer
from twisted.internet.endpoints import serverFromString, clientFromString
from twisted.web.http import HTTPChannel
from twisted.web.static import Data
from twisted.web.resource import Resource
from twisted.web.server import Site

from carml import util
import txtorcon
from txtorcon import TCPHiddenServiceEndpoint


class _PasteBinHTTPChannel(HTTPChannel):
    def connectionMade(self):
        HTTPChannel.connectionMade(self)
        self.site._got_client()

    def connectionLost(self, reason):
        HTTPChannel.connectionLost(self, reason)
        self.site._lost_client()


class PasteBinSite(Site):
    """
    See https://github.com/habnabit/polecat/blob/master/polecat.py for
    the inspriation behind this.

    This class exists so we can count active connections and support a
    command-line option for serving a particular number of
    requests. We need to wait until pending data is written on any
    valid connections that are still active when we reach our limit.
    """

    protocol = _PasteBinHTTPChannel

    def __init__(self, *args, **kw):
        self.active_clients = 0
        self.active_requests = list()
        self._max_requests = kw['max_requests']
        del kw['max_requests']
        self._request_count = 0
        self._stopping_deferred = None
        Site.__init__(self, *args, **kw)

    def getResourceFor(self, request):
        "Override Site so we can track active requests"
        if request.requestHeaders.hasHeader('user-agent'):
            ua = ' '.join(request.requestHeaders.getRawHeaders('user-agent'))
            print('{}: Serving request to User-Agent "{}".'.format(time.asctime(), ua))
        else:
            print('{}: Serving request with no incoming User-Agent header.'.format(time.asctime()))

        # track requsts currently being serviced, so we can nicely
        # shut them off
        self.active_requests.append(request)
        request.notifyFinish().addBoth(self._forget_request, request)

        # see if we've reached the maximum requests
        self._request_count += 1
        if self._max_requests is not None:
            if self._request_count >= self._max_requests:
                d = self.gracefully_stop()
                d.addBoth(lambda x: reactor.stop())

        # call through to parent
        return Site.getResourceFor(self, request)

    def _forget_request(self, request, _):
        self.active_requests.remove(request)

    def _got_client(self):
        self.active_clients += 1

    def _lost_client(self):
        self.active_clients -= 1
        if self.active_clients <= 0 and self._stopping_deferred:
            self._stopping_deferred.callback(None)
            self._stopping_deferred = None

    def gracefully_stop(self):
        "Returns a Deferred that fires when all clients have disconnected."
        if not self.active_clients:
            return defer.succeed(None)
        for request in self.active_requests:
            request.setHeader('connection', 'close')
        self._stopping_deferred = defer.Deferred()
        return self._stopping_deferred


def _progress(percent, tag, message):
    print(util.pretty_progress(percent), message)


@defer.inlineCallbacks
def run(reactor, cfg, tor, dry_run, once, file, count, keys):

    to_share = file.read()
    file.close()

    # stealth auth. keys
    authenticators = []
    if keys:
        for x in xrange(keys):
            authenticators.append('carml_%d' % x)

    if len(authenticators):
        print(len(to_share), "bytes to share with",
              len(authenticators), "authenticated clients.")
    else:
        print(len(to_share), "bytes to share.")
    sys.stdout.flush()

    if dry_run:
        print('Not launching a Tor, listening on 8899.')
        ep = serverFromString(reactor, 'tcp:8899:interface=127.0.0.1')
    elif True:  # connection is None:
        print("Launching Tor.")
        ep = TCPHiddenServiceEndpoint.global_tor(reactor, 80)
        txtorcon.IProgressProvider(ep).add_progress_listener(_progress)
        if keys:
            ep.stealth_auth = [
                'user_{}'.format(n)
                for n in range(keys)
            ]
    else:
        config = yield txtorcon.TorConfig.from_connection(connection)
        ep = txtorcon.TCPEphemeralHiddenServiceEndpoint(reactor, config, 80)

    root = Resource()
    data = Data(to_share, 'text/plain')
    root.putChild('', data)

    if once:
        count = 1
    port = yield ep.listen(PasteBinSite(root, max_requests=count))

    if keys == 0:
        clients = None
    else:
        # FIXME
        clients = port.tor_config.hiddenservices[0].clients

    host = port.getHost()
    if dry_run:
        print("Try it locally via http://127.0.0.1:8899")

    elif clients:
        print("You requested stealth authentication.")
        print("Tor has created %d keys; each key should be given to one person." % len(clients))
        print('They can set one using the "HidServAuth" torrc option, like so:')
        print("")
        for client in clients:
            print("  HidServAuth %s %s" % (client[0], client[1]))
        print("")
        print("Alternatively, any Twisted endpoint-aware client can be given")
        print("the following string as an endpoint:")
        print("")
        for client in clients:
            print("  tor:%s:authCookie=%s" % (client[0], client[1]))
        print("")
        print("For example, using carml:")
        print("")
        for client in clients:
            print("  carml copybin --service tor:%s:authCookie=%s" % (client[0], client[1]))

    else:
        print("People using Tor Browser Bundle can find your paste at (once the descriptor uploads):")
        print("\n   http://{0}\n".format(host.onion_uri))
        print("for example:")
        print("   torsocks curl -o data.asc http://{0}\n".format(host.onion_uri))
        if not count:
            print("Type Control-C to stop serving and shut down the Tor we launched.")
        print("If you wish to keep the hidden-service keys, they're in (until we shut down):")
        print(ep.hidden_service_dir)

    reactor.addSystemEventTrigger('before', 'shutdown',
                                  lambda: print(util.colors.red('Shutting down.')))
    # we never callback() on this, so we serve forever
    d = defer.Deferred()
    yield d



from __future__ import print_function

import os
import sys
import functools

import zope.interface
from twisted.python import usage, log, failure
from twisted.internet import defer, reactor, error
import humanize

import txtorcon

from twisted.internet.defer import inlineCallbacks, Deferred

from carml.util import dump_circuits, format_net_location, nice_router_name, colors, wrap

LOG_LEVELS = ["DEBUG", "INFO", "NOTICE", "WARN", "ERR"]


class BandwidthTracker(object):
    '''
    This tracks bandwidth usage.
    '''

    def __init__(self, maxscale, state):
        #: a list of tuples
        self._bandwidth = []
        self._max = float(maxscale)
        self._state = state

    def circuits(self):
        return len(self._state.circuits)

    def streams(self):
        return len(self._state.streams)

    def on_bandwidth(self, s):
        r, w = map(int, s.split())
        self._bandwidth.append((r, w))
        try:
            self.draw_bars()
        except Exception as e:
            print("bad {}".format(e))

    def on_stream_bandwidth(self, s):
        pass

    def draw_bars(self):
        up = min(1.0, self._bandwidth[-1][0] / self._max)
        dn = min(1.0, self._bandwidth[-1][1] / self._max)
        kbup = self._bandwidth[-1][1] / 1024.0
        kbdn = self._bandwidth[-1][0] / 1024.0

        status = ' ' + colors.green('%.2f' % kbdn)
        status += '/'
        status += colors.red('%.2f' % kbup)  # + ' KiB write'
        status += ' KiB/s'
        status += ' (%d streams, %d circuits)' % (self.streams(), self.circuits())

        # include the paths of any currently-active streams
        streams = ''
        for stream in self._state.streams.values():
            # ...there's a window during which it may not be attached yet
            if stream.circuit:
                circpath = '>'.join(map(lambda r: r.location.countrycode, stream.circuit.path))
                streams += ' ' + circpath
        if len(streams) > 24:
            streams = streams[:21] + '...'
        print(left_bar(up, 20) + unichr(0x21f5) + right_bar(dn, 20) + status + streams)


def left_bar(percent, width):
    '''
    Creates the green/left bar from a percentage and width. It uses
    some unicode vertical-bar characters to get some extra precision
    on the bar-length.
    '''
    blocks = int(percent * width)
    remain = (percent * width) - blocks

    part = int(remain * 8)
    rpart = unichr(0x258f - 7 + part)  # for smooth bar

    return (' ' * (width - blocks)) + colors.negative(colors.green(rpart)) + colors.green(('+' * (blocks)), bg='green')


def right_bar(percent, width):
    '''
    See left_bar(); inverse and in red instead.
    '''
    blocks = int(percent * width)
    remain = (percent * width) - blocks

    part = int(remain * 8)
    rpart = unichr(0x258f - part)  # for smooth bar
    if part == 0:
        rpart = ' '

    return colors.red('+' * (blocks), bg='red') + (colors.red(rpart)) + (' ' * (width - blocks))


@inlineCallbacks
def run(reactor, cfg, tor, max):
    state = yield tor.create_state()
    bwtracker = BandwidthTracker(max, state)
    yield tor.protocol.add_event_listener('BW', bwtracker.on_bandwidth)
    yield tor.protocol.add_event_listener('STREAM_BW', bwtracker.on_stream_bandwidth)

    # infinite loop
    yield Deferred()



from __future__ import print_function


import os
import sys
import functools

import zope.interface
from twisted.python import usage, log, failure
from twisted.internet import defer, reactor, error
import humanize

import txtorcon

from carml.util import dump_circuits, format_net_location, nice_router_name, colors, wrap

LOG_LEVELS = ["DEBUG", "INFO", "NOTICE", "WARN", "ERR"]


def string_for_circuit(state, circuit):
    # path = '->'.join(map(lambda x: x.location.countrycode or '??', circuit.path))
    path = '->'.join(map(lambda x: x.name, circuit.path))
    state = circuit.state
    if state.lower() == 'failed':
        state = colors.red(state)
    return 'Circuit %d (%s) is %s for purpose "%s"' % (circuit.id, path, state, circuit.purpose)


def string_for_stream(state, stream):
    circ = ''
    if stream.circuit:
        circ = ' via circuit %d' % stream.circuit.id
    proc = txtorcon.util.process_from_address(stream.source_addr, stream.source_port, state)
    if proc:
        proc = ' from process "%s"' % (colors.bold(os.path.realpath('/proc/%d/exe' % proc)), )

    elif stream.source_addr == '(Tor_internal)':
        proc = ' for Tor internal use'

    elif stream.source_addr:
        proc = ' from remote "%s:%s"' % (str(stream.source_addr), str(stream.source_port))

    else:
        proc = ''
    hoststring = colors.bold('%s:%d' % (stream.target_host, stream.target_port))
    return 'Stream %d to %s %s%s%s' % (stream.id, hoststring, colors.green('attached'), circ, proc)


class StreamLogger(txtorcon.StreamListenerMixin):
    def __init__(self, state, verbose):
        self.state = state
        self.verbose = verbose

    def stream_attach(self, stream, circuit):
        print(string_for_stream(self.state, stream))
        if self.verbose:
            m = "  " + '->'.join(map(lambda x: nice_router_name(x), circuit.path))
            m += ' (%s)' % ' '.join(map(lambda r: str(r.location.countrycode), circuit.path))
            print(m)

    def stream_failed(self, stream, remote_reason='', **kw):
        print('Stream %d %s because "%s"' % (stream.id, colors.red('failed'),
                                             colors.red(remote_reason)))


def flags(d):
    """
    Basically converts a dict into name=value string, but only if the
    key is all-uppercase. This is because of the way we re-map keyword
    args for the circuit listeners, etcetra.
    """

    r = ''
    for (k, v) in d.iteritems():
        if k.upper() == k:
            r += '%s=%s ' % (k, v)
    return r


class CircuitLogger(txtorcon.CircuitListenerMixin):
    def __init__(self, state, show_flags=False):
        self.state = state
        self.show_flags = show_flags

    def circuit_launched(self, circuit):
        print(string_for_circuit(self.state, circuit))

    def circuit_extend(self, circuit, router):
        print(string_for_circuit(self.state, circuit))

    def circuit_built(self, circuit):
        print(string_for_circuit(self.state, circuit))
        if self.show_flags:
            flagslist = ['%s=%s' % x for x in circuit.flags.items()]
            flags = wrap(' '.join(flagslist), 72, '    ')
            print(colors.cyan('    Flags:'), flags.lstrip())

    def circuit_failed(self, circuit, **kw):
        print('Circuit %d failed (%s).' % (circuit.id, flags(kw)))

    def circuit_closed(self, circuit, **kw):
        print('Circuit %d %s lasted %s (%s).' % (circuit.id,
                                                 colors.red('closed'),
                                                 humanize.time.naturaldelta(circuit.age()),
                                                 flags(kw)))


@zope.interface.implementer(txtorcon.interface.IAddrListener)
class AddressLogger(object):

    def addrmap_added(self, addr):
        print('New address mapping: "%s" -> "%s".' % (addr.name, addr.ip))

    def addrmap_expired(self, name):
        print('Address mapping for "%s" expired.' % name)


def tor_log(level, msg):
    print('%s: %s' % (level, msg))


@defer.inlineCallbacks
def run(reactor, cfg, tor, verbose, no_guards, no_addr, no_circuits, no_streams, once, log_level):
    state = yield tor.create_state()

    follow_string = None
    if log_level and not once:
        follow_string = 'Logging ('
        for event in log_level:  # LOG_LEVELS:
            state.protocol.add_event_listener(event, functools.partial(tor_log, event))
            follow_string += event + ', '
            if event == log_level:
                break
        follow_string = follow_string[:-2] + ')'
    if not no_streams:
        if follow_string:
            follow_string += ' and Stream'
        else:
            follow_string = 'Stream'
        if len(state.streams):
            print("Current streams:")
            for stream in state.streams.values():
                print('  ' + string_for_stream(state, stream))
        else:
            print("No streams.")
        state.add_stream_listener(StreamLogger(state, verbose))

    if not no_circuits:
        if follow_string:
            follow_string += ' and Circuit'
        else:
            follow_string = 'Circuit'

        if len(state.circuits):
            print("Current circuits:")
            dump_circuits(state, verbose=verbose)
        else:
            print("No circuits.")
        state.add_circuit_listener(CircuitLogger(state, show_flags=verbose))

    if not no_guards:
        if len(state.entry_guards):
            print("Current Entry Guards:")
            for (name, router) in state.entry_guards.iteritems():
                if not router.from_consensus:
                    if router.name:
                        print("  %s: %s (not in consensus)" % (router.name, router.id_hex))
                    else:
                        print("  %s (not in consensus)" % router.id_hex)

                else:
                    print(" ", router.id_hex, router.name, format_net_location(router.location))

        else:
            print("No Guard nodes!")

    if not no_addr:
        if follow_string:
            follow_string += ' and Address'
        else:
            follow_string = 'Address'

        if len(state.addrmap.addr):
            print("Current address mappings:")
            for addr in state.addrmap.addr.values():
                print('  %s -> %s' % (addr.name, addr.ip))
            state.addrmap.add_listener(AddressLogger())
        else:
            print("No address mappings.")

    all_done = defer.Deferred()
    if not once:
        print('')
        print("Following new %s activity:" % follow_string)

        def stop_reactor(arg):
            print("Tor disconnected.")
            all_done.callback(None)

        def error(fail):
            print(colors.red('Error:'), fail.getErrorMessage())
        state.protocol.on_disconnect.addErrback(error).addBoth(stop_reactor)

    else:
        all_done.callback(None)
    yield all_done



#!/usr/bin/env python

# putting this in __main__ allows you to "run" the module; it's a
# magic Python thing.
# e.g., like: python -m carml

import sys
from carml import cli

if __name__ == "__main__":
    cli.carml()



from __future__ import print_function

import sys
import datetime
import functools
import random
from collections import defaultdict

from twisted.python import usage, log
from twisted.internet import defer, reactor, endpoints
import zope.interface
import humanize
import txtorcon
from txtorcon.router import hashFromHexId

from carml import util


@defer.inlineCallbacks
def _print_router_info(router, agent=None):
    # loc = yield router.get_location()
    loc = yield router.location
    print(u"            name: {}".format(router.name))
    print(u"          hex id: {}".format(router.id_hex))
    print(u"id hash (base64): {}".format(hashFromHexId(router.id_hex)))
    print(u"        location: {}".format("unknown" if loc.countrycode is None else loc.countrycode))
    print(u"         address: {}:{} (DirPort={})".format(router.ip, router.or_port, router.dir_port))
    print(u"           flags: {}".format(" ".join(router.flags)))
    diff = datetime.datetime.utcnow() - router.modified
    print(u"  last published: {} ago ({})".format(humanize.naturaldelta(diff), router.modified))
    if agent:
        print(util.colors.italic("Extended information from" + util.colors.green(" onionoo.torproject.org") + ":"))
        details = yield router.get_onionoo_details(agent)
        details.setdefault('dir_address', '<none>')
        details.setdefault('city_name', 'unknown')
        details.setdefault('region_name', 'unknown')
        details.setdefault('country_name', 'unknown')
        details['or_addresses'] = ', '.join(details.get('or_addresses', []))
        print(
            u"        platform: {platform}\n"
            u"        runnning: {running}\n"
            u"     dir_address: {dir_address}\n"
            u"    OR addresses: {or_addresses}\n"
            u"        location: {city_name}, {region_name}, {country_name}\n"
            u"       host name: {host_name}\n"
            u"              AS: {as_number} ({as_name})\n"
            u"  last restarted: {last_restarted}\n"
            u"    last changed: {last_changed_address_or_port}\n"
            u"       last seen: {last_seen}\n"
            u"   probabilities: guard={guard_probability} middle={middle_probability} exit={exit_probability}\n"
            u"".format(**details)
        )


@defer.inlineCallbacks
def router_info(state, arg, tor):
    for fp in arg:
        if len(fp) == 40 and not fp.startswith('$'):
            fp = '${}'.format(fp)

        try:
            relay = state.routers[fp]
        except KeyError:
            candidates = [
                r for r in state.all_routers
                if fp in r.name or fp in r.id_hex
            ]
            if not candidates:
                print("Nothing found ({} routers total)".format(len(state.all_routers)))
            if len(candidates) > 1:
                print("Found multiple routers:")
            for router in candidates:
                yield _print_router_info(router)
                print()
        else:
            yield _print_router_info(relay, agent=tor.web_agent())


def _when_updated(state):
    d = defer.Deferred()

    def _newconsensus(doc):
        # we actually don't care what's *in* the event, we just know
        # that the state has now updated...maybe a .when_updated() in
        # TorState?
        print("Got NEWCONSENSUS at {}".format(datetime.datetime.now()))
        d.callback(None)
        state.protocol.remove_event_listener('NEWCONSENSUS', _newconsensus)
    state.protocol.add_event_listener('NEWCONSENSUS', _newconsensus)
    return d


@defer.inlineCallbacks
def _await_router(state, router_id):
    print("Waiting for relay {}".format(router_id))
    while True:
        yield _when_updated(state)
        print("received update")
        try:
            defer.returnValue(state.routers[router_id])
            return
        except KeyError:
            print("{} not found; waiting".format(router_id))
            continue


@defer.inlineCallbacks
def router_await(state, arg):
    if len(arg) == 40 and not arg.startswith('$'):
        arg = '${}'.format(arg)
    if not arg.startswith('$') and len(arg) == 41:
        print("Doesn't appear to be a hex router ID")
        return

    if arg in state.routers:
        print("Router already present:")
        r = state.routers[arg]
    else:
        r = yield _await_router(state, arg)
    yield _print_router_info(r)


def router_list(state):
    for router in state.all_routers:
        print("{}".format(router.id_hex[1:]))


@defer.inlineCallbacks
def run(reactor, cfg, tor, list, info, await):
    state = yield tor.create_state()
    if info:
        yield router_info(state, info, tor)
    elif list:
        yield router_list(state)
    elif await:
        yield router_await(state, await)



from __future__ import print_function
import sys
import time
import functools

from zope.interface import implementer
from twisted.python import usage, log
from twisted.plugin import IPlugin
from twisted.internet import defer
from twisted.web.iweb import IAgentEndpointFactory
from twisted.web.client import Agent, readBody
from twisted.internet import reactor
from twisted.internet.protocol import Protocol


from carml import util
import txtorcon
from txtorcon import TCPHiddenServiceEndpoint


@defer.inlineCallbacks
def run(reactor, cfg, tor, service):
    config = yield tor.get_config()
    socks = yield tor.protocol.get_info('net/listeners/socks')
    socks = socks['net/listeners/socks']
    socks_host, socks_port = socks.split(':')

    args = service.split(':')
    onion = args[1]
    cookie = args[2].split('=')[1]

    auth = '%s %s' % (onion, cookie)
    if auth not in config.HidServAuth:
        config.HidServAuth.append(auth)
    yield config.save()

    agent = tor.web_agent()
    res = yield agent.request('GET', 'http://{}/'.format(onion))
    print('Response: "{} {}" with {} bytes'.format(res.code, res.phrase, res.length))
    data = yield readBody(res)
    print(data)



from __future__ import print_function

import sys
import functools

from twisted.internet import defer


def newid_got_signal(all_done, x):
    if x == 'NEWNYM':
        print('success.')
        all_done.callback(None)


def newid_no_signal(reactor, left, all_done):
    if left > 0:
        print('  waiting {} more seconds.'.format(left))
        reactor.callLater(1, newid_no_signal, reactor, left - 1, all_done)
    else:
        all_done.errback(RuntimeError('no acknowledgement in 10 seconds.'))


@defer.inlineCallbacks
def run(reactor, cfg, tor):
    all_done = defer.Deferred()

    # Tor emits this event whenever it processes a SIGNAL command.
    tor.protocol.add_event_listener('SIGNAL', functools.partial(newid_got_signal, all_done))

    print("Requesting new identity",)
    sys.stdout.flush()
    yield tor.protocol.signal('NEWNYM')
    # answer will be "OK" since Tor received the signal
    # if rate-limiting is happening, the "SIGNAL NEWNYM" event
    # will not be received. so we wait for it
    reactor.callLater(1, newid_no_signal, reactor, 10, all_done)
    yield all_done



# check Python Package Index (PyPI) over 3 different circuits and
# compare sha256 hashes. suitable for use in requirements.txt with
# "twine"
from __future__ import print_function

from zope.interface import implementer
from twisted.python import usage
from twisted.plugin import IPlugin
from twisted.internet import reactor
from twisted.internet import endpoints
from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import deferLater
from twisted.web.client import readBody

import json
import hashlib
from base64 import urlsafe_b64encode
from distutils.version import LooseVersion

import txtorcon
import txtorcon.socks


@inlineCallbacks
def run(reactor, cfg, tor, package_name, package_version):
    agent = tor.web_agent()
    state = yield tor.create_state()

    # download metadata from PyPI over "any" circuit
    uri = b'https://pypi.python.org/pypi/{}/json'.format(package_name)
    print("downloading: '{}'".format(uri))
    resp = yield agent.request(b'GET', uri)
    data = yield readBody(resp)
    data = json.loads(data)

    # did we get a valid sdist URL somewhere?
    sdist_url = None
    version = None
    if package_version is None:
        # print data['releases'].keys()
        available = [LooseVersion(x) for x in data['releases'].keys()]
        package_version = str(sorted(available)[-1])
        print("Using version: {}".format(package_version))

    for url in data['releases'][package_version]:
        if url['packagetype'] == 'sdist':
            sdist_url = url['url'].encode('UTF8')
            version = url['filename']
    if sdist_url is None:
        print("Error: couldn't find any 'sdist' URLs")
        raise RuntimeError("No sdist URL")
    else:
        print("Found sdist: {} for {}".format(sdist_url, version))

    # download the distribution over several different circuits,
    # and record the sha256 hash each time.
    digests = []
    while len(digests) < 3:
        circ = yield state.build_circuit()
        try:
            yield circ.when_built()
        except Exception:
            print("Circuit failed; trying another.")
            continue

        print(
            "Built circuit: {}".format(
                ' -> '.join([r.ip for r in circ.path]),
            )
        )

        try:
            agent = circ.web_agent(reactor, tor._default_socks_endpoint())
            resp = yield agent.request(b'GET', sdist_url)
            # FIXME could stream this to the hasher with a custom
            # protocol, but teh RAMz they are cheap
            tarball = yield readBody(resp)
        except txtorcon.socks.TtlExpiredError as e:
            print("Timed out: {}".format(e))
            continue
        except Exception as e:
            print("Something went wrong: {}".format(e))
            continue

        hasher = hashlib.new('sha256')
        hasher.update(tarball)
        # the whole point is to match peep's hashes, and this is
        # exactly what it does:
        digest = urlsafe_b64encode(hasher.digest()).decode('ascii').rstrip('=')
        digests.append((sdist_url, circ, digest))
        print("sha256:", digest)

    print("Found hashes:")
    feel_fear = False
    for (url, circ, digest) in digests:
        print(digest)
        if digest != digests[0][-1]:
            print("Fearsome Warning! Mismatched digest!!")
            feel_fear = True
            print("Circuit:", '->'.join(map(lambda r: r.hex_id, circ.path)))

    if feel_fear:
        print("****\n  Something fishy!\n****")
    else:
        print("Add this to requirements.txt for peep:")
        print()
        print("# sha256: {}".format(digests[0][-1]))
        print("{}=={}".format(package_name, package_version))
    return



from __future__ import print_function

import os
import sys
import functools

from twisted.python import usage, log
from twisted.internet import defer, reactor
from zope.interface import implementer
import txtorcon
import humanize

from carml import util


names = {
    # onion.debian.org
    '10years.debconf.org': 'b5tearqs4v4nvbup.onion',
    'appstream.debian.org': '5j7saze5byfqccf3.onion',
    'apt.buildd.debian.org': 'ito4xpoj3re4wctm.onion',
    'backports.debian.org': '6f6ejaiiixypfqaf.onion',
    'bits.debian.org': '4ypuji3wwrg5zoxm.onion',
    'blends.debian.org': 'bcwpy5wca456u7tz.onion',
    'bootstrap.debian.net': 'ihdhoeoovbtgutfm.onion',
    'cdimage-search.debian.org': '4zhlmuhqvjkvspwb.onion',
    'd-i.debian.org': 'f6syxyjdgzbeacry.onion',
    'debaday.debian.net': 'ammd7ytxcpeavif2.onion',
    'debconf0.debconf.org': 'ynr7muu3263jikep.onion',
    'debconf1.debconf.org': '4do6yq4iwstidagh.onion',
    'debconf16.debconf.org': '6nhxqcogfcwqzgnm.onion',
    'debconf2.debconf.org': 'ugw3zjsayleoamaz.onion',
    'debconf3.debconf.org': 'zdfsyv3rubuhpql3.onion',
    'debconf4.debconf.org': 'eeblrw5eh2is36az.onion',
    'debconf5.debconf.org': '3m2tlhjsoxws2akz.onion',
    'debconf6.debconf.org': 'gmi5gld3uk5ozvrv.onion',
    'debconf7.debconf.org': '465rf3c2oskkqc24.onion',
    'debdeltas.debian.net': 'vral2uljb3ndhhxr.onion',
    'debug.mirrors.debian.org': 'ktqxbqrhg5ai2c7f.onion',
    'dpl.debian.org': 'j73wbfpplklpixbh.onion',
    'dsa.debian.org': 'f7bphdxlqca3sevt.onion',
    'es.debconf.org': 'nwvk3svshonwqfjs.onion',
    'fr.debconf.org': 'ythg247lqkx2gpgx.onion',
    'ftp.debian.org': 'vwakviie2ienjx6t.onion',
    'ftp.ports.debian.org': 'nbybwh4atabu6xq3.onion',
    'incoming.debian.org': 'oscbw3h7wrfxqi4m.onion',
    'incoming.ports.debian.org': 'vyrxto4jsgoxvilf.onion',
    'lintian.debian.org': 'ohusanrieoxsxlmh.onion',
    'manpages.debian.org': 'pugljpwjhbiagkrn.onion',
    'metadata.ftp-master.debian.org': 'cmgvqnxjoiqthvrc.onion',
    'micronews.debian.org': 'n7jzk5wpel4tdog2.onion',
    'miniconf10.debconf.org': 'tpez4zz5a4civ6ew.onion',
    'mirror-master.debian.org': 'kyk55bof3hzdiwrm.onion',
    'mozilla.debian.net': 'fkbjngvraoici6k7.onion',
    'news.debian.net': 'tz4732fxpkehod36.onion',
    'onion.debian.org': '5nca3wxl33tzlzj5.onion',
    'people.debian.org': 'hd37oiauf5uoz7gg.onion',
    'planet.debian.org': 'gnvweaoe2xzjqldu.onion',
    'release.debian.org': '6nvqpgx7bih375fx.onion',
    'rtc.debian.org': 'ex4gh7cig5ssn2xm.onion',
    'security-team.debian.org': 'ynvs3km32u33agwq.onion',
    'security.debian.org': 'sgvtcaew4bxjd7ln.onion',
    'timeline.debian.net': 'qqvyib4j3fz66nuc.onion',
    'tracker.debian.org': '2qlvvvnhqyda2ahd.onion',
    'wnpp-by-tags.debian.net': 'gl3n4wtekbfaubye.onion',
    'www.debian.org': 'sejnfjrq6szgca7v.onion',
    'www.ports.debian.org': 'lljrzrimek6if67j.onion',

    # onion.torproject.org
    'archive.torproject.org': 'yjuwkcxlgo7f7o6s.onion',
    'atlas.testnet.torproject.org': '2d5quh2deowe4kpd.onion',
    'atlas.torproject.org': '52g5y5karruvc7bz.onion',
    'aus1.torproject.org': 'x3nelbld33llasqv.onion',
    'aus2.torproject.org': 'vijs2fmpd72nbqok.onion',
    'bridges.torproject.org': 'z5tfsnikzulwicxs.onion',
    'cloud.torproject.org': 'icxe4yp32mq6gm6n.onion',
    'collector.testnet.torproject.org': 'vhbbidwvzwhahsrg.onion',
    'collector.torproject.org': 'qigcb4g4xxbh5ho6.onion',
    'collector2.torproject.org': 'kkvj4mhsttfcrksj.onion',
    'compass.torproject.org': 'lwygejoa6fm26eef.onion',
    'consensus-health.torproject.org': 'tgnv2pssfumdedyw.onion',
    'crm.torproject.org': 'sgs4q3dzv74f723x.onion',
    'deb.torproject.org': 'sdscoq7snqtznauu.onion',
    'dist.torproject.org': 'rqef5a5mebgq46y5.onion',
    'donate.torproject.org': 'bjk3o77eebkax2ud.onion',
    'exonerator.torproject.org': 'zfu7x4fuagirknhb.onion',
    'extra.torproject.org': 'klbl4glo2btuwyok.onion',
    'gettor.torproject.org': 'tngjm3owsslo3wgo.onion',
    'git.torproject.org': 'dccbbv6cooddgcrq.onion',
    'gitweb.torproject.org': 'jqs44zhtxl2uo6gk.onion',
    'health.testnet.torproject.org': 'fr6scuhdp5dqvy7d.onion',
    'help.torproject.org': '54nujbl4qohb5qdp.onion',
    'jenkins.torproject.org': 'f7lqb5oicvsahone.onion',
    'media.torproject.org': 'n46o4uxsej2icp5l.onion',
    'metrics.torproject.org': 'rougmnvswfsmd4dq.onion',
    'munin.torproject.org': 'hhr6fex2giwmolct.onion',
    'nagios.torproject.org': 'kakxayzmcc3zeomu.onion',
    'nyx.torproject.org': 'ebxqgaz3dwywcoxl.onion',
    'onion.torproject.org': 'yz7lpwfhhzcdyc5y.onion',
    'onionoo.torproject.org': 'tgel7v4rpcllsrk2.onion',
    'onionperf.torproject.org': 'llhb3u5h3q66ha62.onion',
    'ooni.torproject.org': 'fqnqc7zix2wblwex.onion',
    'people.torproject.org': 'sbe5fi5cka5l3fqe.onion',
    'research.torproject.org': 'wcgqzqyfi7a6iu62.onion',
    'spec.torproject.org': 's2bweojt5vg52e5i.onion',
    'staging.crm.torproject.org': 'swnwd5bhvjk4dd5o.onion',
    'staging.donate.torproject.org': 'cvtwbn7mgxki7gvc.onion',
    'stem.torproject.org': 'vt5hknv6sblkgf22.onion',
    'tb-manual.torproject.org': 'dgvdmophvhunawds.onion',
    'test-data.tbb.torproject.org': 'fylvgu5r6gcdadeo.onion',
    'test.crm.torproject.org': 'abp7hndzgazze2wy.onion',
    'test.donate.torproject.org': 'p73stlm5nhogxw4w.onion',
    'testnet.torproject.org': 'bo7uextohjpuqvrh.onion',
    'trac.torproject.org': 'ea5faa5po25cf7fb.onion',
    'webstats.torproject.org': 'gbinixxw7gnsh5jr.onion',
    'www-staging.torproject.org': 'krkzagd5yo4bvypt.onion',
    'www.onion-router.net': 'hzmun3rnnxjhkyhg.onion',
    'www.torproject.org': 'expyuzz4wqqyqhjn.onion',
}

@implementer(txtorcon.IStreamAttacher)
class Attacher(object):

    def __init__(self, tor):
        self._tor = tor

    @defer.inlineCallbacks
    def attach_stream(self, stream, circuits):
        try:
            remap = names[stream.target_host]
        except KeyError:
            remap = none

        if remap is not None and remap != stream.target_host:
            print("  {} becomes {}".format(stream.target_host, remap))
            cmd = 'REDIRECTSTREAM {} {}'.format(stream.id, remap)
            yield self._tor.protocol.queue_command(cmd)
        defer.returnValue(None)  # ask Tor to attach the stream, always


@defer.inlineCallbacks
def run(reactor, cfg, tor):
    state = yield tor.create_state()
    state.set_attacher(Attacher(tor), reactor)
    yield defer.Deferred()



from __future__ import print_function

import os
import sys
import time
import functools

import zope.interface
from twisted.python import usage, log
from twisted.internet import defer, reactor

import txtorcon

from carml.util import dump_circuits
from carml.util import format_net_location
from carml.util import nice_router_name
from carml.util import colors

import click


@defer.inlineCallbacks
def run(reactor, cfg, tor, list_events, once, show_event, count, events):
    all_events = yield tor.protocol.get_info('events/names')
    all_events = all_events['events/names']
    if list_events:
        for e in all_events.split():
            click.echo(e)
        return

    all_done = defer.Deferred()
    counter = [count]
    if once:
        counter[0] = 1

    def _got_event(evt, msg):
        if counter[0] is not None:
            counter[0] -= 1
            if counter[0] >= 0:
                print(msg)
            elif all_done:
                all_done.callback(None)
                all_done = None
        elif evt:
            print("{}: {}".format(evt, msg))
        else:
            ts = time.asctime()
            print("{} {}".format(ts, msg))
        sys.stdout.flush()

    for e in events:
        e = e.upper()
        if e not in all_events:
            print("Invalid event:", e)
            return
        if show_event:
            listener = functools.partial(_got_event, e)
        else:
            listener = functools.partial(_got_event, None)
        tor.protocol.add_event_listener(e, listener)

    # might be forever if there's no count
    yield all_done



from __future__ import print_function

import sys
import datetime
import functools
import random

from twisted.python import usage, log
from twisted.internet import defer, reactor
import zope.interface
import humanize
import txtorcon

from carml import util


class CircOptions(usage.Options):
    def __init__(self):
        super(CircOptions, self).__init__()
        self.longOpt.remove('version')
        self.longOpt.remove('help')
        self['delete'] = []

    optFlags = [
        ('list', 'L', 'List existing circuits.'),
        ('if-unused', 'u', 'When deleting, pass the IfUnused flag to Tor.'),
        ('verbose', None, 'More information per circuit.'),
    ]

    optParameters = [
        ('build', 'b', '', 'Build a new circuit, given a comma-separated list of router names or IDs. Use "auto" to let Tor select the route.', str),
    ]

    def opt_delete(self, arg):
        """Delete a circuit by its ID."""
        for x in arg.split(','):
            self['delete'].append(int(x))


@defer.inlineCallbacks
def list_circuits(reactor, cfg, tor, verbose):
    print("Circuits:")
    state = yield tor.create_state()

    now = datetime.datetime.utcnow()
    util.dump_circuits(state, verbose)


@defer.inlineCallbacks
def delete_circuit(reactor, cfg, tor, circid, ifunused):
    unused_string = '(if unused) ' if ifunused else ''
    print('Deleting circuit %s"%s"...' % (unused_string, circid),)

    state = yield tor.create_state()  # bootstrap=False)

    kw = {}
    if ifunused:
        kw['IfUnused'] = True

    try:
        circ = state.circuits[circid]
    except KeyError:
        raise RuntimeError("No such circuit '{}'".format(circid))

    status = yield state.close_circuit(circid, **kw)
    print(status, '(waiting for CLOSED)...')
    yield circ.when_closed()
    # we're now awaiting a callback via CIRC events indicating
    # that our circuit has entered state CLOSED


class _BuiltCircuitListener(txtorcon.CircuitListenerMixin):
    def __init__(self, circid, all_done):
        self.circid = circid
        self.first = True
        self._all_done = all_done

    def circuit_extend(self, circuit, router):
        if circuit.id == self.circid:
            if self.first:
                self.first = False
            else:
                sys.stdout.write(' -> ')
            sys.stdout.write(router.name)
            sys.stdout.flush()

    def circuit_built(self, circuit):
        if circuit.id == self.circid:
            print(": " + util.colors.green("built."))
            self._all_done.callback(None)

    def circuit_closed(self, circuit, **kw):
        if circuit.id == self.circid:
            print(": " + util.colors.red("closed."))
            self._all_done.callback(None)

    def circuit_failed(self, circuit, **kw):
        if circuit.id == self.circid:
            r = kw['reason'] if 'reason' in kw else ''
            rr = kw['remote_reason'] if 'remote_reason' in kw else ''
            msg = util.colors.red('failed') + ' (%s, %s).' % (r, rr)
            self._all_done.errback(RuntimeError(msg))


@defer.inlineCallbacks
def build_circuit(reactor, cfg, tor, routers):
    state = yield tor.create_state()

    if len(routers) == 1 and routers[0].lower() == 'auto':
        routers = None
        # print("Building new circuit, letting Tor select the path.")
    else:
        def find_router(args):
            position, name = args
            if name == '*':
                if position == 0:
                    return random.choice(state.entry_guards.values())
                else:
                    return random.choice(state.routers.values())
            r = state.routers.get(name) or state.routers.get('$' + name)
            if r is None:
                if len(name) == 40:
                    print("Couldn't look up %s, but it looks like an ID" % name)
                    r = name
                else:
                    raise RuntimeError('Couldn\'t find router "%s".' % name)
            return r
        routers = map(find_router, enumerate(routers))
        print("Building circuit:", '->'.join(map(util.nice_router_name, routers)))

    try:
        circ = yield state.build_circuit(routers)
        all_done = defer.Deferred()

        sys.stdout.write("Circuit ID %d: " % circ.id)
        sys.stdout.flush()
        state.add_circuit_listener(_BuiltCircuitListener(circ.id, all_done))
        # all_done will callback when the circuit is built (or errback
        # if it fails).

    except txtorcon.TorProtocolError as e:
        log.err(e)

    yield all_done


@defer.inlineCallbacks
def run(reactor, cfg, tor, if_unused, verbose, list, build, delete):
    if list:
        yield list_circuits(reactor, cfg, tor, verbose)

    elif len(delete) > 0:
        deletes = []
        for d in delete:
            deletes.append(delete_circuit(reactor, cfg, tor, d, if_unused))
        results = yield defer.DeferredList(deletes)
        for ok, value in results:
            if not ok:
                raise value

    elif build:
        yield build_circuit(reactor, cfg, tor, build.split(','))



from __future__ import print_function
import sys
import time
import functools

import zope.interface
from twisted.python import usage, log
from twisted.plugin import IPlugin
from twisted.internet import reactor
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.internet.endpoints import serverFromString
from twisted.web.http import HTTPChannel
from twisted.web.static import Data
from twisted.web.resource import Resource
from twisted.web.server import Site

from carml import util
import txtorcon
from txtorcon import TCPHiddenServiceEndpoint


@inlineCallbacks
def run(reactor, cfg, tor, ports):

    def info(msg):
        if 'Service descriptor (v2) stored' in msg:
            got_upload.callback(None)
    got_upload = Deferred()
    tor.protocol.add_event_listener('INFO', info)

    hs = txtorcon.EphemeralHiddenService(ports)
    yield hs.add_to_tor(tor.protocol)
    print("Created HS", hs.hostname)

    def remove():
        print("removing hidden-service")
        return hs.remove_from_tor(tor.protocol)
    reactor.addSystemEventTrigger('before', 'shutdown', remove)

    print("...waiting for descriptor upload")
    yield got_upload
    print("Got one.", time.asctime())

    # we never callback() on this, so we serve forever
    d = Deferred()
    yield d


def _remove_service(config, hs):
    config.hiddenservices.remove(hs)
    return config.save()



#
# FIXME: better name? "the command called cmd" is pretty weird
#

from __future__ import print_function

import os
import sys
import functools

import zope.interface
from twisted.python import usage, log
from twisted.protocols.basic import LineReceiver
from twisted.internet import defer, reactor, stdio

import txtorcon
from carml import util


class StdioLineReceiver(LineReceiver):
    delimiter = '\n'

    def __init__(self, all_done, proto):
        self.proto = proto
        self.all_done = all_done
        self.outstanding = []
        self._exit = False      # when True, all_done.callback() when outstanding empties

    def connectionMade(self):
        print("Keep entering keys to run CMD on. Control-d to exit.", file=sys.stderr)

    def lineReceived(self, line):
        # Ignore blank lines
        if not line:
            return

        keys = line.split()
        d = do_cmd(self.proto, tuple(keys))
        d.addCallback(self._completed, d)
        d.addErrback(self._error, d)
        self.outstanding.append(d)

    def _error(self, arg, d):
        print(util.colors.red(str(arg)))
        return

    def _completed(self, arg, d):
        self.outstanding.remove(d)
        if self._exit and len(self.outstanding) == 0:
            self.all_done.callback(None)

    def connectionLost(self, reason):
        if len(self.outstanding):
            self._exit = True
        else:
            self.all_done.callback(None)


def do_cmd(proto, args):
    def _print(res):
        print(res)

    def _error(arg):
        print(util.colors.red(arg.getErrorMessage()))
        return None
    d = proto.queue_command(' '.join(args))
    d.addCallback(_print)
    d.addErrback(_error)
    return d


# see cmd_info for an alternate way to implement this via a method
# with attributes and "zope.interface.implementsDirectly()"
# trying out both ways to see what feels better
@defer.inlineCallbacks
def run(reactor, cfg, tor, args):
    if len(args) == 0:
        print("(no command to run)")

    elif len(args) == 1 and args[0] == '-':
        all_done = defer.Deferred()
        stdio.StandardIO(StdioLineReceiver(all_done, tor.protocol))
        yield all_done
    else:
        yield do_cmd(tor.protocol, args)



from __future__ import print_function

import pkg_resources
import click


# @defer.inlineCallbacks
def run(reactor, cfg, tor):
    readme = pkg_resources.resource_string('carml', '../README.rst')
    # uhm, docutils documentation is confusing as all hell and no good
    # examples of "convert this rST string to anything else" .. :/ but
    # we should "render" it to text
    click.echo_via_pager(readme)



import sys
import os
import shutil
import re
from setuptools import setup, find_packages


__version__ = '18.0.0'


setup(
    name='carml',
    version=__version__,
    author='meejah',
    author_email='meejah@meejah.ca',
    url='https://github.com/meejah/carml',
    license='Public Domain (http://unlicense.org/)',
    description='A command-line tool to query and control a running Tor. Based on txtorcon + Twisted.',
    long_description=open('README.rst', 'r').read(),
    keywords=['python', 'twisted', 'tor', 'command-line', 'cli'],
    install_requires=[
        'humanize',
        'ansicolors',
        'backports.lzma',
        'txtorcon>=0.19.3',
        'pyasn1',
        'click>=6.7',
    ],
    classifiers=[
        'Framework :: Twisted',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: Public Domain',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Networking',
        'Topic :: Internet :: Proxy Servers',
        'Topic :: Internet',
        'Topic :: Security',
        'Topic :: Utilities'],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'carml = carml.cli:carml'
        ]
    },
    include_package_data=True,
    package_data={'': ['*.asc', '*.pem']},
    data_files=[
        ('share/carml', ['README.rst', 'meejah.asc']),
        ('share/carml/doc/', ['doc/' + x for x in filter(lambda x: x.endswith('.rst'), os.listdir('doc'))]),
    ]
)



from __future__ import print_function

import os
import sys
import functools

from twisted.python import usage, log
from twisted.internet import defer, reactor
from zope.interface import implementer
import txtorcon
import humanize

from carml import util


def attach_streams_per_process(state):
    print("Exiting (e.g. Ctrl-C) will cause Tor to resume choosing circuits.")
    print("Giving each new PID we see its own Circuit (until they're gone).")

    @implementer(txtorcon.IStreamAttacher)
    class Attacher(object):

        def __init__(self):
            self.pid_to_circuits = {}

        def choose_new_circuit(self, stream, circuits):
            for circ in circuits.values():
                if circ in self.pid_to_circuits.values():
                    continue
                if circ.state != 'BUILT':
                    continue
                return circ
            raise RuntimeError("Ran out of circuits to select.")

        def attach_stream(self, stream, circuits):
            src_addr, src_port = stream.flags['SOURCE_ADDR'].split(':')
            pid = txtorcon.util.process_from_address(src_addr, src_port)
            procname = os.path.realpath('/proc/%d/exe' % pid)

            try:
                circ = self.pid_to_circuits[pid]
            except KeyError:
                circ = self.choose_new_circuit(stream, circuits)
                self.pid_to_circuits[pid] = circ
                print('Selected circuit %d for process %d (%s).' % (circ.id, pid, procname))
                print('  ', '->'.join([p.name if p.name_is_unique else ('{%s}' % p.name) for p in circ.path]))

#            if stream.state == 'NEWRESOLVE':
#                print "  attaching %d (resolve %s)" % (stream.id, stream.target_host)
#            else:
            print("  attaching stream %d to circuit %d for %s:%d (%s)" % (stream.id, circ.id, stream.target_host, stream.target_port, procname))
            return circ

    state.set_attacher(Attacher(), reactor)


def attach_streams_to_circuit(circid, state):
    try:
        circ = state.circuits[circid]
    except KeyError:
        print("Circuit {} doesn't exist.".format(circid))
        return None
    print("Exiting (e.g. Ctrl-C) will cause Tor to resume choosing circuits.")
    print("Attaching all new streams to Circuit %d." % circ.id)
    print("   ", '->'.join([p.name if p.name_is_unique else ('~%s' % p.name) for p in circ.path]))

    @implementer(txtorcon.IStreamAttacher)
    class Attacher(txtorcon.CircuitListenerMixin):

        def circuit_closed(self, this_circ, **kw):
            if circ == this_circ:
                print("Circuit {} vanished (REASON={}, REMOTE_REASON={})".format(
                    circ.id,
                    kw.get('REASON', 'not specified'),
                    kw.get('REMOTE_REASON', 'not specified'),
                ))
                # should we just exit now?
                # Pro: kind-of makes sense
                # Con: if you're expecting streams to go via "your"
                # circuit, maybe you want them to "fail closed" and
                # not work at all -- which is what I'm doing right now
                # so -> probably want exiting to be an option, and not the default

        def attach_stream(self, stream, circuits):
            if stream.flags.get('PURPOSE', 'unknown') in ['DIR_FETCH', 'DIR_UPLOAD', 'DIRPORT_TEST']:
                print("  tor-internal directory stream ({})".format(stream.flags['PURPOSE']))
                return None
            if circ.state == 'CLOSED':
                print("  target circuit is closed, not attaching")
                return txtorcon.TorState.DO_NOT_ATTACH
            if stream.state == 'NEWRESOLVE':
                print("  attaching %d (resolve %s)" % (stream.id, stream.target_host))
            else:
                print("  attaching %d %s:%d" % (stream.id, stream.target_host,
                                                stream.target_port))
            return circ

    attacher = Attacher()
    state.set_attacher(attacher, reactor)
    state.add_circuit_listener(attacher)
    # FIXME doesn't exit on control-c? :(
    d = defer.Deferred()
    d.addBoth(lambda x: print('foo', x))
    return d


def list_streams(state, verbose):
    print("Streams:")
    for stream in state.streams.values():
        flags = str(stream.flags) if stream.flags else 'no flags'
        state = stream.state
        state_to_color = dict(SUCCEEDED=util.colors.green,
                              FAILED=util.colors.red)
        if state in state_to_color:
            state = state_to_color[state](state)
        print("  %d: %s on circuit %d (%s)" % (stream.id, state, stream.circuit.id,
                                               flags))
        if verbose:
            h = stream.target_addr if stream.target_addr else stream.target_host
            source = txtorcon.util.process_from_address(stream.source_addr, stream.source_port)
            if source is None:
                source = 'unknown'
            print("     to %s:%s, from %s" % (h, stream.target_port, source))


@defer.inlineCallbacks
def close_stream(state, streamid):
    class DetermineStreamClosure(object):
        def __init__(self, target_id, done_d):
            self.circ_id = str(target_id)
            self.stream_gone = False
            self.already_deleted = False
            self.completed_d = done_d

        def __call__(self, text):
            cid, what, _ = text.split(' ', 2)
            if what in ['CLOSED', 'FAILED']:
                if self.circ_id == cid:
                    self.stream_gone = True
                    print("gone (%s)..." % self.circ_id,)
                    sys.stdout.flush()
                    if self.already_deleted:
                        self.completed_d.callback(self)
    if streamid not in state.streams:
        print('No such stream "%s".' % streamid)
        return
    print('Closing stream "%s"...' % (streamid, ))

    gone_d = defer.Deferred()
    monitor = DetermineStreamClosure(streamid, gone_d)
    state.protocol.add_event_listener('STREAM', monitor)
    sys.stdout.flush()

    try:
        status = yield state.streams[streamid].close()
        status = status.state
        monitor.already_deleted = True
    except txtorcon.TorProtocolError as e:
        print(util.colors.red('Error: ') + e.what())
        return

    if monitor.stream_gone:
        print(status)
        return

    print('%s (waiting for CLOSED)...' % status)
    sys.stdout.flush()
    yield gone_d
    # we're now awaiting a callback via CIRC events indicating
    # that our stream has entered state CLOSED


class StreamBandwidth(object):
    """
    The bandwidth-events of a single stream
    """
    #  __slots__ = ['_events']

    def __init__(self, max_live=20, roll_up=5):
        self._events = []  # list of 3-tuples
        self._history = []
        self.max_live = max_live
        self.roll_up = roll_up

        # XXX we could recursively roll-up too, i.e. spill from one
        # bucket to the next. but, for now, there are precisely two
        # bucks: the "current", and the first one containing (avg,
        # min, max) etc

    def add_bandwidth(self, epoch, read, write):
        self._events.append((epoch, read, write))
        self._maybe_truncate()

    def _maybe_truncate(self):
        """
        If we've gone past our max_live amount by at least roll_up events,
        we push it into the history (possibly also truncating that).
        """
        # XXX should we examine seconds here instead? i.e. have a
        # max-seconds (instead of going by event-count)?
        if len(self._events) > self.max_live + self.roll_up:
            rolling = self._events[:self.roll_up]
            self._events = self._events[self.roll_up:]
            duration = float(self._events[0][0] - rolling[0][0])
            mean_r = sum(x[1] for x in rolling) / duration
            mean_w = sum(x[2] for x in rolling) / duration

            # XXX should append some smarter-er object instead of tuple?
            # (start, duration, mean_r, mean_w, max_r, max_w)
            self._history.append(
                (
                    rolling[0][0],
                    duration,
                    mean_r, mean_w,
                    max(x[1] for x in rolling),
                    max(x[2] for x in rolling),
                )
            )
            self._history = self._history[-10:]
            print("HISTORY NOW", self._history)
            print("age {}, total bw {}".format(
                self._events[-1][0] - (self._history[-1][0] + self._history[-1][1]),
                sum([sum(x[1:]) for x in self._history]),
            ))

    def bytes_read(self):
        return sum(event[1] for event in self._events)

    def bytes_written(self):
        return sum(event[2] for event in self._events)

    def duration(self):
        if not self._events:
            return 0.0
        if len(self._events) == 1:
            return 1.0
        return float(self._events[-1][0] - self._events[0][0]) + 1.0

    def rate(self):
        span = self.duration()
        if span == 0.0:
            return (0.0, 0.0)  # mmm...pragmatism
        return (self.bytes_read() / span, self.bytes_written() / span)


class BandwidthMonitor(txtorcon.StreamListenerMixin):
    @staticmethod
    @defer.inlineCallbacks
    def create(reactor, state):
        bw = BandwidthMonitor(reactor, state)
        yield bw._setup()
        defer.returnValue(bw)

    def __init__(self, reactor, state):
        self._reactor = reactor  # just IReactorClock required?
        self._state = state
        self._active = {}  # maps stream ID -> list-of-tuples

    def stream_new(self, stream):
        print("new", stream)
        self._active[stream.id] = StreamBandwidth()

    def stream_succeeded(self, stream):
        # i think this happens when it *starts* passing data?
        print("succeeded", stream, stream.target_host, stream.target_addr)

    def stream_attach(self, stream, circuit):
        pass

    def stream_detach(self, stream, **kw):
        pass

    def stream_closed(self, stream, **kw):
        # print("closed", stream, self._active)
        if stream.id not in self._active:
            print(
                "Previously unknown stream to {stream.target_host} died".format(
                    stream=stream,
                )
            )
        else:
            bw = self._active[stream.id]
            print(
                "Stream {stream.id} to {stream.target_host}: {read} read, {written} written in {duration:.1f}s ({read_rate})".format(
                    stream=stream,
                    read=util.colors.green(humanize.naturalsize(bw.bytes_read())),
                    written=util.colors.red(humanize.naturalsize(bw.bytes_written())),
                    read_rate=humanize.naturalsize(sum(bw.rate())) + '/s',
                    duration=bw.duration(),
                )
            )

    def stream_failed(self, stream, **kw):
        pass

    def _stream_bw(self, bw):
        sid, wr, rd = [int(x) for x in bw.split()]
        try:
            bandwidth = self._active[sid]
        except KeyError:
            bandwidth = self._active[sid] = StreamBandwidth()
        bandwidth.add_bandwidth(self._reactor.seconds(), rd, wr)

    @defer.inlineCallbacks
    def _setup(self):
        yield self._state.add_stream_listener(self)
        yield self._state.protocol.add_event_listener('STREAM_BW', self._stream_bw)


@defer.inlineCallbacks
def monitor_streams(state, verbose):
    print("monitor", state, verbose)
    from twisted.internet import reactor
    bw = yield BandwidthMonitor.create(reactor, state)


@defer.inlineCallbacks
def run(reactor, cfg, tor, list, follow, attach, close, verbose):
    state = yield tor.create_state()
    if attach:
        yield attach_streams_to_circuit(attach, state)
    elif list:
        yield list_streams(state, verbose)
    elif close:
        yield close_stream(state, close)
    elif follow:
        d = defer.succeed(None)
        yield monitor_streams(state, verbose)
        yield defer.Deferred()



from __future__ import absolute_import, print_function

import sys

from twisted.internet import defer, task
from twisted.internet.endpoints import clientFromString
from twisted.python.failure import Failure
from twisted.python import usage, log

import click
import txtorcon

from . import carml_readme
from . import carml_check_pypi
from . import carml_stream
from . import carml_events
from . import carml_circ
from . import carml_cmd
from . import carml_monitor
from . import carml_newid
from . import carml_pastebin
from . import carml_copybin
from . import carml_relay
from . import carml_tbb
from . import carml_temphs
from . import carml_tmux
from . import carml_xplanet
from . import carml_graph


LOG_LEVELS = ["DEBUG", "INFO", "NOTICE", "WARN", "ERR"]


class LogObserver(object):
    def __init__(self, timestamp=False, flush=True):
        self.timestamp = timestamp
        self.flush = flush
        # we keep our own copies of these, because Twisted's
        # startLoggingWithObserver overwrites them with its own
        # monitoring machinery
        self.stdout = sys.stdout
        self.stderr = sys.stderr

    def __call__(self, arg):
        # we don't want to print out every little thing logged by
        # Twisted or txtorcon, just what we output (which is always
        # print()-ed)
        try:
            if not arg['printed']:
                return
        except KeyError:
            return
        msg = ' '.join(arg['message'])

        # possibly add timestamps
        if self.timestamp:
            msg = util.colors.cyan(time.asctime()) + ': ' + msg

        # figure out if we want stdout or stderr
        out = self.stdout
        if 'isError' in arg and arg['isError']:
            out = self.stderr
            if not msg and 'failure' in arg:
                msg = util.colors.red('Error: ') + arg['failure'].getErrorMessage()

        # actually print message
        print(msg, file=out)
        if self.flush:
            out.flush()


class Config(object):
    '''
    Passed as the Click object (@pass_obj) to all CLI methods.
    '''


@click.group()
@click.option('--timestamps', '-t', help='Prepend timestamps to each line.', is_flag=True)
@click.option('--no-color', '-n', help='Same as --color=no.', is_flag=True, default=None)
@click.option('--info', '-i', help='Show version of Tor we connect to (on stderr).', is_flag=True)
@click.option('--quiet', '-q', help='Some commands show less information with this option.', is_flag=True)
@click.option('--debug', '-d', help='Debug; print stack traces on error.', is_flag=True)
@click.option(
    '--password', '-p',
    help=('Password to authenticate to Tor with. Using cookie-based authentication'
          'is much easier if you are on the same machine.'),
    default=None,
    required=False,
)
@click.option(
    '--connect', '-c',
    default='tcp:host=127.0.0.1:port=9051',
    help=('Where to connect to Tor. This accepts any Twisted client endpoint '
          'string, or an ip:port pair. Examples: "tcp:localhost:9151" or '
          '"unix:/var/run/tor/control".'),
    metavar='ENDPOINT',
)
@click.option(
    '--color', '-C',
    type=click.Choice(['auto', 'no', 'always']),
    default='auto',
    help='Colourize output using ANSI commands.',
)
@click.pass_context
def carml(ctx, timestamps, no_color, info, quiet, debug, password, connect, color):
    if (color == 'always' and no_color) or \
       (color == 'no' and no_color is True):
        raise click.UsageError(
            "--no-color={} but --color={}".format(no_color, color)
        )

    cfg = Config()
    ctx.obj = cfg

    cfg.timestamps = timestamps
    cfg.no_color = no_color
    cfg.info = info
    cfg.quiet = quiet
    cfg.debug = debug
    cfg.password = password
    cfg.connect = connect
    cfg.color = color

    # start logging
    _log_observer = LogObserver()
    log.startLoggingWithObserver(_log_observer, setStdout=False)


def _run_command(cmd, cfg, *args, **kwargs):

    @defer.inlineCallbacks
    def _startup(reactor):
        ep = clientFromString(reactor, cfg.connect)
        tor = yield txtorcon.connect(reactor, ep)

        if cfg.info:
            info = yield tor.proto.get_info('version', 'status/version/current', 'dormant')
            click.echo(
                'Connected to a Tor version "{version}" (status: '
                '{status/version/current}).\n'.format(**info)
            )
        yield defer.maybeDeferred(
            cmd, reactor, cfg, tor, *args, **kwargs
        )

    from twisted.internet import reactor
    codes = [0]

    def _the_bad_stuff(f):
        print("Error: {}".format(f.value))
        if cfg.debug:
            print(f.getTraceback())
        codes[0] = 1
        return None

    def _go():
        d = _startup(reactor)
        d.addErrback(_the_bad_stuff)
        d.addBoth(lambda _: reactor.stop())

    reactor.callWhenRunning(_go)
    reactor.run()
    sys.exit(codes[0])


@carml.command()
@click.pass_obj
def readme(cfg):
    """
    Show the README.rst
    """
    return _run_command(
        carml_readme.run,
        cfg,
    )


@carml.command()
@click.option(
    '--package', '-p',
    help='Name of the package to check (unfortunately, case matters)',
    required=True,
)
@click.option(
    '--revision', '-r',
    help='Specific version to check (default: latest)',
    default=None,
)
@click.pass_obj
def check_pypi(cfg, package, revision):
    """
    Check a PyPI package hash across multiple circuits.
    """
    return _run_command(
        carml_check_pypi.run,
        cfg, package, revision,
    )


@carml.command()
@click.option(
    '--if-unused', '-u',
    help='When deleting, pass the IfUnused flag to Tor.',
    is_flag=True,
)
@click.option(
    '--verbose',
    help='More information per circuit.',
    is_flag=True,
)
@click.option(
    '--list', '-L',
    help='List existing circuits.',
    is_flag=True,
    default=None,
)
@click.option(
    '--build', '-b',
    help=('Build a new circuit, given a comma-separated list of router names or'
          ' IDs. Use "auto" to let Tor select the route.'),
    default=None,
)
@click.option(
    '--delete',
    help='Delete a circuit by its ID.',
    default=None,
    multiple=True,
    type=int,
)
@click.pass_obj
def circ(cfg, if_unused, verbose, list, build, delete):
    """
    Manipulate Tor circuits.
    """
    if len([o for o in [list, build, delete] if o]) != 1:
        raise click.UsageError(
            "Specify just one of --list, --build or --delete"
        )
    return _run_command(
        carml_circ.run,
        cfg, if_unused, verbose, list, build, delete,
    )


@carml.command()
@click.argument(
    "command_args",
    nargs=-1,
)
@click.pass_obj
def cmd(cfg, command_args):
    """
    Run the rest of the args as a Tor control command. For example
    "GETCONF SocksPort" or "GETINFO net/listeners/socks".
    """
    return _run_command(
        carml_cmd.run,
        cfg, command_args,
    )


@carml.command()
@click.option(
    '--list', '-L',
    help='Show available events.',
    is_flag=True,
)
@click.option(
    '--once',
    help='Output exactly one and quit (same as -n 1 or --count=1).',
    is_flag=True,
)
@click.option(
    '--show-event', '-s',
    help='Prefix each line with the event it is from.',
    is_flag=True,
)
@click.option(
    '--count', '-n',
    help='Output this many events, and quit (default is unlimited).',
    type=int,
)
@click.argument(
    "events",
    nargs=-1,
)
@click.pass_obj
def events(cfg, list, once, show_event, count, events):
    """
    Follow any Tor events, listed as positional arguments.
    """
    if len(events) < 1 and not list:
        raise click.UsageError(
            "Must specify at least one event"
        )
    return _run_command(
        carml_events.run,
        cfg, list, once, show_event, count, events,
    )


@carml.command()
@click.option(
    '--list', '-L',
    help='List existing streams.',
    is_flag=True,
)
@click.option(
    '--follow', '-f',
    help='Follow stream creation.',
    is_flag=True,
)
@click.option(
    '--attach', '-a',
    help='Attach all new streams to a particular circuit-id.',
    type=int,
    default=None,
)
@click.option(
    '--close', '-d',
    help='Delete/close a stream by its ID.',
    type=int,
    default=None,
)
@click.option(
    '--verbose', '-v',
    help='Show more details.',
    is_flag=True,
)
@click.pass_context
def stream(ctx, list, follow, attach, close, verbose):
    """
    Manipulate Tor streams.
    """
    cfg = ctx.obj
    if len([x for x in [list, follow, attach, close] if x]) != 1:
        click.echo(ctx.get_help())
        raise click.UsageError(
            "Must specify one of --list, --follow, --attach or --close"
        )
    return _run_command(
        carml_stream.run,
        cfg, list, follow, attach, close, verbose,
    )


@carml.command()
@click.option(
    '--once', '-o',
    help='Exit after printing the current state.',
    is_flag=True,
)
@click.option(
    '--no-streams', '-s',
    help='Without this, list Tor streams.',
    is_flag=True,
)
@click.option(
    '--no-circuits', '-c',
    help='Without this, list Tor circuits.',
    is_flag=True,
)
@click.option(
    '--no-addr', '-a',
    help='Without this, list address mappings (and expirations, with -f).',
    is_flag=True,
)
@click.option(
    '--no-guards', '-g',
    help='Without this, Information about your current Guards.',
    is_flag=True,
)
@click.option(
    '--verbose', '-v',
    help='Additional information. Circuits: ip, location, asn, country-code.',
    is_flag=True,
)
@click.option(
    '--log-level', '-l',
    default=[],
    type=click.Choice(LOG_LEVELS),
    multiple=True,
)
@click.pass_context
def monitor(ctx, verbose, no_guards, no_addr, no_circuits, no_streams, once, log_level):
    """
    General information about a running Tor; streams, circuits,
    address-maps and event monitoring.
    """
    cfg = ctx.obj
    return _run_command(
        carml_monitor.run,
        cfg, verbose, no_guards, no_addr, no_circuits, no_streams, once, log_level,
    )


@carml.command()
@click.pass_context
def newid(ctx):
    """
    Ask Tor for a new identity via NEWNYM, and listen for the response
    acknowledgement.
    """
    cfg = ctx.obj
    return _run_command(
        carml_newid.run,
        cfg,
    )


@carml.command()
@click.option(
    '--dry-run', '-d',
    help='Test locally; no Tor launch.',
    is_flag=True,
)
@click.option(
    '--once',
    help='Same as --count=1.',
    is_flag=True,
)
@click.option(
    '--file', '-f',
    default=sys.stdin,
    type=click.File('rb'),
    help='Filename to use as input (instead of stdin)',
)
@click.option(
    '--count', '-n',
    default=None,
    help='Number of requests to serve.',
    type=int,
)
@click.option(
    '--keys', '-k',
    default=0,
    help='Number of authentication keys to create.',
    type=int,
)
@click.pass_context
def pastebin(ctx, dry_run, once, file, count, keys):
    """
    Put stdin (or a file) on a fresh hidden-service easily.
    """
    if count is not None and count < 0:
        raise click.UsageError(
            "--count must be positive"
        )
    if once and count is not None:
        raise click.UsageError(
            "Only specify one of --count or --once"
        )

    cfg = ctx.obj
    return _run_command(
        carml_pastebin.run,
        cfg, dry_run, once, file, count, keys,
    )


@carml.command()
@click.option(
    '--list',
    help='List all relays by hex ID.',
    is_flag=True,
)
@click.option(
    '--info',
    default='',
    help='Look up by fingerprint (or part of one).',
)
@click.option(
    '--info-file',
    default=None,
    type=click.File('r'),
    help='Look up multiple fingerprints, one per line from the given file',
)
@click.option(
    '--await',
    default='',
    help='Monitor NEWCONSENSUS for a fingerprint to exist',
)
@click.pass_context
def relay(ctx, list, info, await, info_file):
    """
    Information about Tor relays.
    """
    if not list and not info and not await and not info_file:
        raise click.UsageError(
            "Require one of --list, --info, --await, --info-file"
        )
    if info_file:
        infos = [info] if info else []
        infos.extend(
            [x.strip() for x in info_file.readlines()]
        )
    else:
        infos = [info] if info else []

    cfg = ctx.obj
    return _run_command(
        carml_relay.run,
        cfg, list, infos, await,
    )


@carml.command()
@click.option(
    '--beta', '-b',
    help='Use the beta release (if available).',
    is_flag=True,
)
@click.option(
    '--alpha', '-a',
    help='Use the alpha release (if available).',
    is_flag=True,
)
#        ('hardened', 'H', 'Use a hardened release (if available).'),
@click.option(
    '--use-clearnet', '-',
    help='Do the download over plain Internet, NOT via Tor (NOT RECOMMENDED).',
    is_flag=True,
)
@click.option(
    '--system-keychain', '-K',
    help='Instead of creating a temporary keychain with provided Tor keys, use the current user\'s existing GnuPG keychain.',
    is_flag=True,
)
@click.option(
    '--no-extract', '-E',
    help='Do not extract after downloading.',
    is_flag=True,
)
@click.option(
    '--no-launch', '-L',
    help='Do not launch TBB after downloading.',
    is_flag=True,
)
@click.pass_context
def tbb(ctx, beta, alpha, use_clearnet, system_keychain, no_extract, no_launch):
    """
    Download the lastest Tor Browser Bundle (with pinned SSL
    certificates) and check the signatures.
    """
    if not no_extract:
        try:
            import backports.lzma
        except ImportError:
            raise click.UsageError(
                'You need "backports.lzma" installed to do 7zip extraction.'
                ' (Pass --no-extract to skip extraction).'
            )

    cfg = ctx.obj
    return _run_command(
        carml_tbb.run,
        cfg, beta, alpha, use_clearnet, system_keychain, no_extract, no_launch,
    )


@carml.command()
@click.option(
    '--port', '-p',
    help='Port to pass-through (or "remote:local" for different local port)',
    multiple=True,
)
@click.pass_context
def temphs(ctx, port):
    """
    Add a temporary hidden-service to the Tor we connect to.

    This keeps a hidden-service running as long as this command is
    running with an arbitrary list of forwarded ports.
    """
    if len(port) == 0:
        raise click.UsageError(
            "Specify at least one port"
        )

    def _range_check(p):
        try:
            p = int(p)
        except ValueError:
            raise click.UsageError(
                "{} is not an int".format(p)
            )

    for p in port:
        if ':' in p:
            remote, local = p.split(':')
            _range_check(remote)
            _range_check(local)
        else:
            _range_check(p)

    cfg = ctx.obj
    return _run_command(
        carml_temphs.run,
        cfg, list(port),
    )


@carml.command()
@click.pass_context
def tmux(ctx):
    """
    Show some informantion in your tmux status.

    Output for tmux's status-right. As in, put this in ~/.tmux.conf::

        set-option -g status-utf8 on
        set-option -g status-fg green
        set -g status-right '#(carml tmux)'
        set -g status-interval 2
    """
    cfg = ctx.obj
    return _run_command(
        carml_tmux.run,
        cfg,
    )


@carml.command()
@click.option(
    '--all', '-A',
    help='Output all the routers, not just your own guards and circuits.',
    is_flag=True,
)
@click.option(
    '--execute', '-x',
    help='Run the xplanet command in a tempdir',
    is_flag=True,
)
@click.option(
    '--follow', '-f',
    help='Implies -x, re-running whenever a new circuit enters BUILT state.',
    is_flag=True,
)
@click.option(
    '--arc-file', '-a',
    help='Also output current circuits in an xplanet "arc_file" compatible file.',
    default=None,
    type=click.File('w'),
)
@click.option(
    '--file',
    help='Filename to dump markers too (default is stdout).',
    default=sys.stdout,
    type=click.File('w'),
)
@click.pass_context
def xplanet(ctx, all, execute, follow, arc_file, file):
    """
    """
    cfg = ctx.obj
    return _run_command(
        carml_xplanet.run,
        cfg, all, execute, follow, arc_file, file,
    )


@carml.command()
@click.option(
    '--service', '-s',
    help='The endpoint you were given to download from (like "tor:xxxxx.onion:authCookie=xxxxxx".',
    default=None,
    required=True,
    metavar='EP',
)
@click.pass_context
def copybin(ctx, service):
    """
    Download something from a "pastebin" hidden-service.
    """
    cfg = ctx.obj
    return _run_command(
        carml_copybin.run,
        cfg, service,
    )


@carml.command()
@click.option(
    '--max', '-m',
    help='Maximum scale, in bytes.',
    default=1024 * 20,
)
@click.pass_context
def graph(ctx, max):
    """
    A nice coloured console bandwidth-graph.
    """
    cfg = ctx.obj
    return _run_command(
        carml_graph.run,
        cfg, max,
    )


@carml.command()
@click.argument(
    'what'
)
@click.pass_context
def help(ctx, what):
    """
    Print help on sub-commands (like "carml help events").
    """
    try:
        cmd = globals()[what]
    except KeyError:
        print('No such command "carml {}".'.format(what))
    else:
        context = click.Context(cmd, ctx.parent, info_name=what)
        fmt = context.make_formatter()
        cmd.format_help(context, fmt)
        for b in fmt.buffer:
            print(b, end='')



from __future__ import print_function

import json
import platform
import os
import re
import time
import subprocess
import shutil
import tempfile
import functools
import pkg_resources
from six import BytesIO

try:
    import OpenSSL
except ImportError:
    raise RuntimeError("No OpenSSL: try 'pip install pyOpenSSL'")

import zope.interface
from twisted.python import usage, log
from twisted.plugin import IPlugin
from twisted.internet import reactor, defer, ssl, endpoints, error
from twisted.internet._sslverify import PublicKey
from twisted.internet.protocol import Protocol
from twisted.web.client import Agent, ProxyAgent, RedirectAgent, ResponseDone, ResponseFailed
from twisted.web.iweb import IPolicyForHTTPS
from twisted.web.http_headers import Headers

from carml import util

DEBUG = False

# TODO:
#
# stream isolation: we should ask for a Tor connection, add a
# SOCKSPort to it for our own use, and connect to that (and then
# delete it from config when done). Unless there's an easier way.

# FIXME
# Twisted 14.0.0 can "just do" chain verification, I believe
# at *least* verify this does a similar thing, or just depend on >=14
# and delete it


@zope.interface.implementer(IPolicyForHTTPS)
class VerifyCertChainContextFactory(ssl.ClientContextFactory):
    def __init__(self, cert_chain):
        '''
        :param cert_chain:
            A list containing ``twisted.internet.ssl.Certificate``
            instances, with the "depth 0" one being first. Create
            these with ``twisted.internet.ssl.Certificate.loadPEM()``
        '''
        self.chain = {}         # key=int (depth), value=certificate
        for (i, cert) in enumerate(cert_chain):
            self.chain[i] = cert
        if DEBUG:
            for i in range(len(self.chain)):
                print('%d: %s' % (i, self.chain[i].getSubject()))

    def creatorForNetloc(self, hostname, port):
        # this should return a "creator" for the given hostname/port
        # -- we always return ourself because we only verify for a
        # single cert-chain so don't care what host we're visiting.
        return self

    def getContext(self):
        ctx = OpenSSL.SSL.Context(OpenSSL.SSL.TLSv1_METHOD)

        # load just the DigiCert root as "the" certificate store

        # WARNING although it's tempting to just call
        # set_default_verify_paths(), on OS X that fails. And
        # apparently if you fail to have a certificate store at all,
        # OpenSSL fails to actually-verify anything, even with all
        # the asking that follows here AND if you just "return 0"
        # from the verify_hostname callback...so don't do that.
        # FIXME there is something we could do here, maybe? can we
        # confirm that the cert store is in such a state that OpenSSL
        # will actually do verifications??!
        store = ctx.get_cert_store()
        store.add_cert(OpenSSL.crypto.load_certificate(OpenSSL.SSL.FILETYPE_PEM,
                                                       pkg_resources.resource_string('carml', 'keys/digicert-root-ca.pem')))

        # FAIL_IF_NO_PEER_CERT is requireCertification in Twisted
        # VERIFY_CLIENT_ONCE is verifyOnce in Twisted
        ctx.set_verify(OpenSSL.SSL.VERIFY_PEER | OpenSSL.SSL.VERIFY_FAIL_IF_NO_PEER_CERT | OpenSSL.SSL.VERIFY_CLIENT_ONCE,
                       self.verify_hostname)

        # verifyDepth from Twisted
        ctx.set_verify_depth(len(self.chain))

        # fixBrokenPeer from Twisted
        if False:
            _OP_ALL = getattr(OpenSSL.SSL, 'OP_ALL', 0x0000FFFF)
            ctx.set_options(_OP_ALL)

        # enableSingleUseKeys in Twisted
        ctx.set_options(OpenSSL.SSL.OP_SINGLE_DH_USE)

        # enableSessions in Twisted
        # see http://twistedmatrix.com/trac/browser/tags/releases/twisted-13.2.0/twisted/internet/_sslverify.py#L800
        # name = "%s-%d" % (reflect.qual(self.__class__), _sessionCounter())
        # sessionName = md5(networkString(name)).hexdigest()

        # _OP_NO_TICKET = 0x00004000
        # could disallow session tickets with ctx.set_options(self._OP_NO_TICKET)
        # if enableSessions in Twisted, calls ctx.set_session_id(md5(string + counter))
        return ctx

    def verify_hostname(self, connection, cert, errno, depth, preverifyOK):
        if DEBUG:
            print('Verify: pre=%d depth=%d cert=%s issuer=%s' % (preverifyOK,
                                                                 depth,
                                                                 cert.get_subject(),
                                                                 cert.get_issuer()))
        if not preverifyOK:
            # FIXME if below is true about constant-time, then this
            # shouldn't bail out early either...
            return False

        if depth >= len(self.chain):
            print("depth is %d, but we have a chain %d entries long." % (depth, len(self.chain)))
            return False

        # FIXME TODO requires thinking
        # need to ensure this is constant-time? e.g. compare hashes?
        # or are we already screwed on that since we'll bail out of
        # this whole "verify" callback stack early when one cert
        # fails?

        # correct way is to compare hashes (then we only have to store
        # the hash of the public cert, not the actual thing)

        verify_pubkey = PublicKey(cert.get_pubkey())
        golden_pubkey = self.chain[depth].getPublicKey()
        if DEBUG:
            print('incoming="%s", golden="%s"' % (verify_pubkey.keyHash(), golden_pubkey.keyHash()))

        if not golden_pubkey.matches(verify_pubkey):
            # getting out the CN (common name) for a nicer output,
            # but maybe we don't want that -- let "experts" examine
            # "failed.pem" and everyone else just needs to know "it
            # didn't work"?
            cn = ''
            for (k, v) in cert.get_subject().get_components():
                if k == 'CN':
                    cn = v
            common_name = ''.join(map(lambda x: str(x[1]), filter(lambda x: x[0] == 'CN', cert.get_subject().get_components())))
            print('Certificate chain verification failed for "%s".' % common_name)
            print('Public key md5 hash is "%s" but wanted "%s".' % (verify_pubkey.keyHash(), golden_pubkey.keyHash()))
            print('Dumping failing certificate to "failed.pem".')
            with open('failed.pem', 'a') as f:
                f.write(OpenSSL.crypto.dump_certificate(OpenSSL.SSL.FILETYPE_PEM, cert))
            return False

        # already checked preverifyOK
        return True


class ResponseReceiver(Protocol):
    def __init__(self, content, total, all_done):
        self.content = content
        self.total = int(total)
        self.current = 0
        self.last_output = None
        self.start = time.time()  # FIXME use IReactorTime to get this!
        self.all_done = all_done

    def dataReceived(self, data):
        self.content.write(data)
        self.update_progress(len(data))

    def update_progress(self, datalen):
        self.current += datalen
        if self.total > 1024 and self.current:
            howmuch = int((self.current * 100) / self.total)
            if self.last_output is None or howmuch - self.last_output >= 10:
                self.last_output = howmuch
                elapsed = time.time() - self.start
                velocity = self.current / elapsed
                cur = self.current
                tot = self.total
                for (order, unit) in [(1024, 'KiB'), (1024 * 1024, 'MiB'), (1024 * 1024 * 1024, 'GiB')]:
                    if tot / order < 100:
                        break
                cur /= order
                tot /= order
                units = unit
                estimate = self.total / velocity
                left = estimate - elapsed
                print('%s - %.1f of %.1f %s (%ds remaining)' % (util.pretty_progress(howmuch, 5), cur, tot, units, left))
                if self.current == self.total:
                    print('%0.2f MiB/s' % (self.total / (1024.0 * 1024) / elapsed))

    def connectionLost(self, reason):
        print("boom {}".format(reason))
        self.all_done.callback(str(reason))


@defer.inlineCallbacks
def download(agent, uri, filelike):
    resp = yield agent.request(b'GET', uri)
    while resp.code == 302:
        newloc = resp.headers.getRawHeaders('location')[0]
        print("Following 302:", newloc)
        resp = yield agent.request(b'GET', newloc.encode('ascii'))

    if resp.code != 200:
        raise RuntimeError('Failed to download "{}": {}'.format(uri, resp.code))
    done = defer.Deferred()
    total = resp.length
    dl = ResponseReceiver(filelike, total, done)
    resp.deliverBody(dl)
    yield done


def get_download_urls(plat, arch, target_version):
    if '64' in arch:
        arch = 64
    else:
        arch = 32
    sig_fname = dict(linux='tor-browser-linux%d-%s_en-US.tar.xz.asc' % (arch, target_version),
                     darwin='TorBrowserBundle-%s-osx32_en-US.zip.asc' % (target_version, ),
                     windows='torbrowser-install-%s_en-US.exe.asc' % (target_version, ),
                     )[plat]

    dist_fname = sig_fname[:-4]
    return (sig_fname, dist_fname)


def extraction_instructions(fname):
    print("To extract:")
    print("   7z x %s" % fname)
    print("   tar xf %s" % fname[:-3])
    print("or just:")
    print('   7z -bd -so e "%s" | tar xf -' % fname)


def extract_7zip(fname):
    import backports.lzma as lzma
    import tarfile
    lz = lzma.open(str(fname))
    print('Extracting "%s"...' % fname)
    print('  decompressing...')
    tar = tarfile.open(fileobj=lz)

    def progress_generator(tar):
        prog = 0
        so_far = 0
        total = len(tar.getmembers())
        last = 0.0
        for ti in tar:
            so_far += 1
            percent = int((float(so_far) / float(total)) * 100.0)
            if last is None or percent - last >= (100.0 / 5.0):
                last = percent
                print('  %3d%% extracted' % percent)
            yield ti
    tar.extractall(members=progress_generator(tar))
    return None


@defer.inlineCallbacks
def run(reactor, cfg, tor, beta, alpha, use_clearnet, system_keychain, no_extract, no_launch):
    # NOTE the middle cert changed on April 10 or thereabouts;
    # still need to confirm this is legitimate?
    chain = [ssl.Certificate.loadPEM(pkg_resources.resource_string('carml', 'keys/torproject.pem')),
             ssl.Certificate.loadPEM(pkg_resources.resource_string('carml', 'keys/digicert-sha2.pem')),
             ssl.Certificate.loadPEM(pkg_resources.resource_string('carml', 'keys/digicert-root-ca.pem')),
             ]
    cf = VerifyCertChainContextFactory(chain)

    if use_clearnet:
        print(util.colors.red('WARNING') + ': downloading over plain Internet (not via Tor).')
        agent = Agent(reactor, contextFactory=cf)

    else:
        agent = tor.web_agent()

    uri = b'https://www.torproject.org/projects/torbrowser/RecommendedTBBVersions'
    data = BytesIO()
    print(u'Getting recommended versions from "{}".'.format(uri.decode('ascii')))
    d = download(agent, uri, data)

    def ssl_errors(fail):
        if hasattr(fail.value, 'reasons'):
            msg = ''.join([str(r.value.args[-1]) for r in fail.value.reasons])
            raise RuntimeError(msg)
        return fail
    d.addErrback(ssl_errors)
    yield d

    # valid platforms from check.torproject.org can be one of:
    # 'Linux', 'MacOS' or 'Windows'
    plat = platform.system().lower()
    arch = platform.uname()[-2]
    plat_to_tor = dict(linux='Linux', darwin='MacOS', windows='Win')
    if plat not in plat_to_tor:
        print('Unknown platform "%s".' % plat)
        raise RuntimeError('Unknown platform "%s".' % plat)
    tor_plat = plat_to_tor[plat]

    data_bytes = data.getvalue()
    try:
        versions = json.loads(data_bytes.decode('utf8'))

    except:
        print('Error getting versions; invalid JSON:')
        print(data_bytes)
        raise RuntimeError('Invalid JSON:\n{}'.format(data_bytes))

    alpha_re = re.compile(r'[0-9]*.[0-9]*a[0-9]-(Windows|MacOS|Linux)')
    beta_re = re.compile(r'[0-9]*.[0-9]*b[0-9]-(Windows|MacOS|Linux)')
    hardened_re = re.compile(r'(.*)-hardened-(.*)')

    print(util.wrap(', '.join(versions), 60, '  '))
    alphas = filter(lambda x: alpha_re.match(x), versions)
    betas = filter(lambda x: beta_re.match(x), versions)
    # the 'hardened' browser names don't follow the pattern of the
    # others; for now, just ignoring them... (XXX FIXME)
    hardened = filter(lambda x: hardened_re.match(x), versions)
    others = set(versions).difference(alphas, betas, hardened)
    if alpha:
        versions = alphas
    elif beta:
        versions = betas
    else:
        versions = others

    if alphas:
        print(util.colors.yellow("Note: there are alpha versions available; use --alpha to download."))
    if betas:
        print(util.colors.yellow("Note: there are beta versions available; use --beta to download."))
    if hardened:
        print(util.colors.yellow("Note: there are hardened versions available but we don't support downloading them yet."))

    target_version = None
    for v in versions:
        if v.endswith(tor_plat):
            target_version = v[:v.rfind('-')]

    if target_version is None:
        print("Can't find a version to download")
        print("          My platform is: %s (%s)" % (plat, plat_to_tor[plat]))
        print("  Potential versions are: %s" % ', '.join(versions))
        if beta:
            print("(Try without --beta)")
        elif alpha:
            print("(Try without --alpha)")
        raise RuntimeError("Nothing to download found.")

    # download the signature, then browser-bundle (if they don't
    # already exist locally).
    sig_fname, dist_fname = get_download_urls(plat, arch, target_version)
    for to_download in [sig_fname, dist_fname]:
        uri = u'https://www.torproject.org/dist/torbrowser/{}/{}'.format(target_version, to_download).encode('ascii')
        if os.path.exists(to_download):
            print(util.colors.red(to_download) + ': already exists, so not downloading.')
        else:
            def cleanup(failure, fname):
                print('removing "%s"...' % fname)
                os.unlink(fname)
                return failure

            f = open(to_download, 'wb')
            print('Downloading "%s".' % to_download)
            d = download(agent, uri, f)
            d.addErrback(cleanup, to_download)
            yield d
            f.close()

    # ensure the signature matches
    if verify_signature(sig_fname, system_gpg=system_keychain):
        print(util.colors.green("Signature is good."))

        if no_extract:
            print("Download and signature check of the Tor Browser Bundle")
            print("has SUCCEEDED.\n")
            print("It is here: %s\n" % os.path.realpath(dist_fname))
            extraction_instructions(dist_fname)
            print("and then:")

        else:
            try:
                extract_7zip(dist_fname)
                print("Tor Browser Bundle downloaded and extracted.")

            except ImportError:
                msg = 'You need "backports.lzma" installed to do 7zip extraction.'
                print(util.colors.red('Error: ') + msg, isError=True)
                extraction_instructions(dist_fname)

        # running instructions
        lang = dist_fname[-12:-7]
        tbb_path = './tor-browser_%s/Browser/start-tor-browser' % lang
        if no_launch:
            print("To run: %s" % tbb_path)
        else:
            print("running: %s" % tbb_path)
            os.execl(tbb_path, tbb_path)

    else:
        print(util.colors.bold('Deleting tarball; signature verification failed.'))
        os.unlink(dist_fname)
        print('...however signature file is being kept for reference (%s).' % sig_fname)


def verify_signature(fname, system_gpg=False):

    verify_command = ['gpg', '--quiet']
    td = None
    status = False              # pessimism!

    try:
        if not system_gpg:
            # create temporary homedir
            td = tempfile.mkdtemp()
            verify_command.extend(['--homedir', td])

            # add Tor project-people keys to it (the ones who
            # sign releases, anyway)
            keys = []
            keys_path = os.path.join(td, 'keys')
            os.mkdir(keys_path)
            for k in pkg_resources.resource_listdir('carml', 'keys'):
                if k.endswith('.asc'):
                    keys.append(pkg_resources.resource_filename('carml', os.path.join('keys', k)))

            if len(keys) == 0:
                raise RuntimeError('Internal error: failed to find shipped keys.')

            try:
                if subprocess.check_call(['gpg', '--quiet', '--homedir', td, '--import'] + keys):
                    raise RuntimeError("Key import failed.")
            except IOError:
                raise RuntimeError("GPG verification failed; is GnuPG installed?")

        verify_command.extend(['--verify', fname])
        try:
            subprocess.check_call(verify_command)
            status = True

        except subprocess.CalledProcessError:
            print(util.colors.bold(util.colors.red("Signature verification failed.")))
            status = False

    finally:
        if td:
            shutil.rmtree(td)
    return status



'''
This is a simple sphinx extension to provide nice output from "click"
decorated command-lines. Enables markup like this in the Sphinx rst
files::

.. carml_command:: downloadbundle

Which will document the command "carml downloadbundle"
'''

from docutils import nodes
from docutils.parsers.rst import Directive


class carml_command(nodes.Element):
    pass


class CarmlCommandDirective(Directive):
    has_content = True
    required_arguments = 1

    def run(self):
        env = self.state.document.settings.env

        node = carml_command()
        node.line = self.lineno
        node['command'] = self.arguments[0]
        return [node]

    # targetid = "cb-cmd-%d" % env.new_serialno('cb-cmd')
    # targetnode = nodes.target('', '', ids=[targetid])
    # return [targetnode, node]


# shitcrackers, these aren't Click commands :/
def document_commands(app, doctree):
    return ''
    for node in doctree.traverse(carml_command):
        cmd = find_command(node.get('command', None))
        section = nodes.section()
        # section.append(nodes.subtitle(text='carml {}'.format(cmd.name)))
        section.append(nodes.paragraph(text=cmd.help))
        params = nodes.bullet_list()
        section.append(params)
        for param in cmd.params:
            item = nodes.list_item()
            arg = '--{}'.format(param.name)
            if param.metavar:
                arg = '--{} {}'.format(param.name, param.metavar)
            item.append(nodes.literal('', arg))
            item.append(nodes.Text(': ' + str(param.help)))
            if param.show_default:
                item.append(nodes.Text(' (default: '))
                item.append(nodes.literal('', str(param.default)))
                item.append(nodes.Text(')'))
            params.append(item)

        node.replace_self(section)


def setup(app):
    app.add_directive('carml_command', CarmlCommandDirective)
    app.connect(str('doctree-read'), document_commands)
